# -*- coding: utf-8 -*-

"""
legit.config
~~~~~~~~~~~~~~~~~~

This module provides the Legit settings feature set.

"""


class Settings(object):
    _singleton = {}

    # attributes with defaults
    __attrs__ = tuple()

    def __init__(self, **kwargs):
        super(Settings, self).__init__()

        self.__dict__ = self._singleton

    def __call__(self, *args, **kwargs):
        # new instance of class to call
        r = self.__class__()

        # cache previous settings for __exit__
        r.__cache = self.__dict__.copy()
        map(self.__cache.setdefault, self.__attrs__)

        # set new settings
        self.__dict__.update(*args, **kwargs)

        return r

    def __enter__(self):
        pass

    def __exit__(self, *args):

        # restore cached copy
        self.__dict__.update(self.__cache.copy())
        del self.__cache

    def __getattribute__(self, key):
        if key in object.__getattribute__(self, '__attrs__'):
            try:
                return object.__getattribute__(self, key)
            except AttributeError:
                return None
        return object.__getattribute__(self, key)


legit_settings = Settings()

legit_settings.config_defaults = (
    ('allow_black_foreground', 'True',
        'Is the epic black foreground color allowed? Defaults to True.'),

    ('disable_colors', 'False',
        'Y U NO FUN? Defaults to False.'),
)


legit_settings.update_url = 'https://api.github.com/repos/frostming/legit/tags'
legit_settings.forbidden_branches = ['HEAD']



# -*- coding: utf-8 -*-

"""
legit.cli
~~~~~~~~~

This module provides the CLI interface to legit.
"""
import os

import click
import crayons
from clint import resources
from clint.textui import columns

from .core import __version__
from .scm import SCMRepo
from .settings import legit_settings
from .utils import (
    black, format_help, git_version, order_manually, output_aliases,
    status_log, verbose_echo, program_path
)

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
pass_scm = click.make_pass_decorator(SCMRepo)


class LegitGroup(click.Group):
    """Custom Group class with specially sorted command list"""

    command_aliases = {
        'pub': 'publish',
        'sw': 'switch',
        'sy': 'sync',
        'unp': 'unpublish',
        'un': 'undo',
    }

    def list_commands(self, ctx):
        """Override for showing commands in particular order"""
        commands = super(LegitGroup, self).list_commands(ctx)
        return [cmd for cmd in order_manually(commands)]

    def get_command(self, ctx, cmd_name):
        """Override to handle command aliases"""
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        cmd_name = self.command_aliases.get(cmd_name, "")
        return click.Group.get_command(self, ctx, cmd_name)

    def get_help_option(self, ctx):
        """Override for showing formatted main help via --help and -h options"""
        help_options = self.get_help_option_names(ctx)
        if not help_options or not self.add_help_option:
            return

        def show_help(ctx, param, value):
            if value and not ctx.resilient_parsing:
                if not ctx.obj:
                    # legit main help
                    click.echo(format_help(ctx.get_help()))
                else:
                    # legit sub-command help
                    click.echo(ctx.get_help(), color=ctx.color)
                ctx.exit()
        return click.Option(
            help_options,
            is_flag=True,
            is_eager=True,
            expose_value=False,
            callback=show_help,
            help='Show this message and exit.')


@click.group(cls=LegitGroup, invoke_without_command=True, context_settings=CONTEXT_SETTINGS)
@click.version_option(prog_name=black('legit', bold=True), version=__version__)
@click.option('--verbose', is_flag=True, help='Enables verbose mode.')
@click.option('--fake', is_flag=True, help='Show but do not invoke git commands.')
@click.option('--install', is_flag=True, help='Install legit git aliases.')
@click.option('--uninstall', is_flag=True, help='Uninstall legit git aliases.')
@click.option('--config', is_flag=True, help='Edit legit configuration file.')
@click.pass_context
def cli(ctx, verbose, fake, install, uninstall, config):
    """legit command line interface"""
    # Create a repo object and remember it as as the context object.  From
    # this point onwards other commands can refer to it by using the
    # @pass_scm decorator.
    ctx.obj = SCMRepo()
    ctx.obj.fake = fake
    ctx.obj.verbose = fake or verbose

    if install:
        do_install(ctx, verbose, fake)
        ctx.exit()
    elif uninstall:
        do_uninstall(ctx, verbose, fake)
        ctx.exit()
    elif config:
        do_edit_settings(fake)
        ctx.exit()
    else:
        if ctx.invoked_subcommand is None:
            # Display help to user if no commands were passed.
            click.echo(format_help(ctx.get_help()))


@cli.command(short_help='Switches to specified branch.')
@click.argument('to_branch', required=False)
@click.option('--verbose', is_flag=True, help='Enables verbose mode.')
@click.option('--fake', is_flag=True, help='Show but do not invoke git commands.')
@pass_scm
def switch(scm, to_branch, verbose, fake):
    """Switches from one branch to another, safely stashing and restoring local changes.
    """
    scm.fake = fake
    scm.verbose = fake or verbose

    scm.repo_check()

    if to_branch is None:
        scm.display_available_branches()
        raise click.BadArgumentUsage('Please specify a branch to switch to')

    scm.stash_log()
    status_log(scm.checkout_branch, 'Switching to {0}.'.format(
        crayons.yellow(to_branch)), to_branch)
    scm.unstash_log()


@cli.command(short_help='Synchronizes the given branch with remote.')
@click.argument('to_branch', required=False)
@click.option('--verbose', is_flag=True, help='Enables verbose mode.')
@click.option('--fake', is_flag=True, help='Show but do not invoke git commands.')
@pass_scm
@click.pass_context
def sync(ctx, scm, to_branch, verbose, fake):
    """Stashes unstaged changes, Fetches remote data, Performs smart
    pull+merge, Pushes local commits up, and Unstashes changes.

    Defaults to current branch.
    """
    scm.fake = fake
    scm.verbose = fake or verbose

    scm.repo_check(require_remote=True)

    if to_branch:
        # Optional branch specifier.
        branch = scm.fuzzy_match_branch(to_branch)
        if branch:
            is_external = True
            original_branch = scm.get_current_branch_name()
        else:
            raise click.BadArgumentUsage(
                "Branch {0} does not exist. Use an existing branch."
                .format(crayons.yellow(branch)))
    else:
        # Sync current branch.
        branch = scm.get_current_branch_name()
        is_external = False

    if branch in scm.get_branch_names(local=False):
        if is_external:
            ctx.invoke(switch, to_branch=branch, verbose=verbose, fake=fake)
        scm.stash_log(sync=True)
        status_log(scm.smart_pull, 'Pulling commits from the server.')
        status_log(scm.push, 'Pushing commits to the server.', branch)
        scm.unstash_log(sync=True)
        if is_external:
            ctx.invoke(switch, to_branch=original_branch, verbose=verbose, fake=fake)
    else:
        raise click.BadArgumentUsage(
            "Branch {0} is not published. Publish before syncing."
            .format(crayons.yellow(branch)))


@cli.command(short_help='Publishes specified branch to the remote.')
@click.argument('to_branch', required=False)
@click.option('--verbose', is_flag=True, help='Enables verbose mode.')
@click.option('--fake', is_flag=True, help='Show but do not invoke git commands.')
@pass_scm
def publish(scm, to_branch, verbose, fake):
    """Pushes an unpublished branch to a remote repository."""
    scm.fake = fake
    scm.verbose = fake or verbose

    scm.repo_check(require_remote=True)
    branch = scm.fuzzy_match_branch(to_branch)

    if not branch:
        branch = scm.get_current_branch_name()
        scm.display_available_branches()
        if to_branch is None:
            click.echo("Using current branch {0}".format(crayons.yellow(branch)))
        else:
            click.echo(
                "Branch {0} not found, using current branch {1}"
                .format(crayons.red(to_branch), crayons.yellow(branch)))

    branch_names = scm.get_branch_names(local=False)

    if branch in branch_names:
        raise click.BadArgumentUsage(
            "Branch {0} is already published. Use a branch that is not published."
            .format(crayons.yellow(branch)))

    status_log(scm.publish_branch, 'Publishing {0}.'.format(
        crayons.yellow(branch)), branch)


@cli.command(short_help='Removes specified branch from the remote.')
@click.argument('published_branch', required=False)
@click.option('--verbose', is_flag=True, help='Enables verbose mode.')
@click.option('--fake', is_flag=True, help='Show but do not invoke git commands.')
@pass_scm
def unpublish(scm, published_branch, verbose, fake):
    """Removes a published branch from the remote repository."""
    scm.fake = fake
    scm.verbose = fake or verbose

    scm.repo_check(require_remote=True)
    branch = scm.fuzzy_match_branch(published_branch)

    if not branch:
        scm.display_available_branches()
        raise click.BadArgumentUsage('Please specify a branch to unpublish')

    branch_names = scm.get_branch_names(local=False)

    if branch not in branch_names:
        raise click.BadArgumentUsage(
            "Branch {0} is not published. Use a branch that is published."
            .format(crayons.yellow(branch)))

    status_log(scm.unpublish_branch, 'Unpublishing {0}.'.format(
        crayons.yellow(branch)), branch)


@cli.command()
@click.option('--verbose', is_flag=True, help='Enables verbose mode.')
@click.option('--fake', is_flag=True, help='Show but do not invoke git commands.')
@click.option('--hard', is_flag=True, help='Discard local changes.')
@pass_scm
def undo(scm, verbose, fake, hard):
    """Removes the last commit from history."""
    scm.fake = fake
    scm.verbose = fake or verbose

    scm.repo_check()

    status_log(scm.undo, 'Last commit removed from history.', hard)


@cli.command()
@click.argument('wildcard_pattern', required=False)
@pass_scm
def branches(scm, wildcard_pattern):
    """Displays a list of branches."""
    scm.repo_check()

    if wildcard_pattern:
        scm.display_available_branches(wildcard_pattern)
    else:
        scm.display_available_branches()


def do_install(ctx, verbose, fake):
    """Installs legit git aliases."""
    click.echo('The following git aliases will be installed:\n')
    aliases = cli.list_commands(ctx)
    if git_version() >= (2, 23, 0):
        click.echo(
            'As git 2.23.0 introduces a new command "git switch", alias "switch"'
            ' will be skipped.\nUse "legit switch" instead.'
        )
        aliases.remove('switch')
    output_aliases(aliases)

    program = program_path()
    if click.confirm('\n{}Install aliases above?'.format('FAKE ' if fake else ''), default=fake):
        for alias in aliases:
            cmd = '!{} {}'.format(program, alias)
            system_command = 'git config --global --replace-all alias.{0} "{1}"'.format(alias, cmd)
            verbose_echo(system_command, verbose, fake)
            if not fake:
                os.system(system_command)
        if not fake:
            click.echo("\nAliases installed.")
    else:
        click.echo("\nAliases will not be installed.")


def do_uninstall(ctx, verbose, fake):
    """Uninstalls legit git aliases, including deprecated legit sub-commands."""
    aliases = cli.list_commands(ctx)
    # Add deprecated aliases
    aliases.extend(['graft', 'harvest', 'sprout', 'resync', 'settings', 'install', 'uninstall'])
    for alias in aliases:
        system_command = 'git config --global --unset-all alias.{0}'.format(alias)
        verbose_echo(system_command, verbose, fake)
        if not fake:
            os.system(system_command)
    if not fake:
        click.echo('\nThe following git aliases are uninstalled:\n')
        output_aliases(aliases)


def do_edit_settings(fake):
    """Opens legit settings in editor."""

    path = resources.user.open('config.ini').name

    click.echo('Legit Settings:\n')

    for (option, _, description) in legit_settings.config_defaults:
        click.echo(columns([crayons.yellow(option), 25], [description, None]))
    click.echo("")  # separate settings info from os output

    if fake:
        click.echo(crayons.red('Faked! >>> edit {}'.format(path)))
    else:
        click.edit(filename=path)


def handle_abort(aborted, type=None):
    click.echo('{0} {1}'.format(crayons.red('Error:'), aborted.message))
    click.echo(str(aborted.log))
    if type == 'merge':
        click.echo('Unfortunately, there was a merge conflict.'
                   ' It has to be merged manually.')
    elif type == 'unpublish':
        click.echo(
            '''It seems that the remote branch is deleted.
            If `legit branches` still shows it as published,
            then probably the branch has been deleted at the remote by someone else.
            You can run `git fetch --prune` to update remote information.
            ''')
    raise click.Abort


legit_settings.abort_handler = handle_abort



# -*- coding: utf-8 -*-

"""
legit.core
~~~~~~~~~~

This module provides the basic functionality of legit.
"""

from . import bootstrap
del bootstrap

__version__ = '1.2.0'
__author__ = 'Kenneth Reitz'
__license__ = 'BSD'



import click
from clint.textui import colored, columns
import crayons
import os
import sys

from .settings import legit_settings


def status_log(func, message, *args, **kwargs):
    """Emits header message, executes a callable, and echoes the return strings."""

    click.echo(message)
    log = func(*args, **kwargs)

    if log:
        out = []

        for line in log.split('\n'):
            if not line.startswith('#'):
                out.append(line)
        click.echo(black('\n'.join(out)))


def verbose_echo(str, verbose=False, fake=False):
    """Selectively output ``str``, with special formatting if ``fake`` is True"""
    verbose = fake or verbose

    if verbose:
        color = crayons.green
        prefix = ''
        if fake:
            color = crayons.red
            prefix = 'Faked!'
        click.echo(color('{} >>> {}'.format(prefix, str)))


def output_aliases(aliases):
    """Display git aliases"""
    for alias in aliases:
        cmd = '!legit ' + alias
        click.echo(columns([colored.yellow('git ' + alias), 20], [cmd, None]))


def order_manually(sub_commands):
    """Order sub-commands for display"""
    order = [
        "switch",
        "sync",
        "publish",
        "unpublish",
        "undo",
        "branches",
    ]
    ordered = []
    commands = dict(zip([cmd for cmd in sub_commands], sub_commands))
    for k in order:
        ordered.append(commands.get(k, ""))
        if k in commands:
            del commands[k]

    # Add commands not present in `order` above
    for k in commands:
        ordered.append(commands[k])

    return ordered


def format_help(help):
    """Format the help string."""
    help = help.replace('Options:', str(black('Options:', bold=True)))

    help = help.replace('Usage: legit', str('Usage: {0}'.format(black('legit', bold=True))))

    help = help.replace('  switch', str(crayons.green('  switch', bold=True)))
    help = help.replace('  sync', str(crayons.green('  sync', bold=True)))
    help = help.replace('  publish', str(crayons.green('  publish', bold=True)))
    help = help.replace('  unpublish', str(crayons.green('  unpublish', bold=True)))
    help = help.replace('  undo', str(crayons.green('  undo', bold=True)))
    help = help.replace('  branches', str(crayons.yellow('  branches', bold=True)))

    additional_help = \
        """Usage Examples:
Switch to specific branch:
$ {0}

Sync current branch with remote:
$ {1}

Sync current code with a specific remote branch:
$ {2}

Publish current branch to remote:
$ {3}

Publish a specific branch to remote:
$ {4}

Unpublish a specific branch from remote:
$ {5}

List branches matching wildcard pattern:
$ {6}

Commands:""".format(
            crayons.red('legit switch <branch>'),
            crayons.red('legit sync'),
            crayons.red('legit sync <branch>'),
            crayons.red('legit publish'),
            crayons.red('legit publish <branch>'),
            crayons.red('legit unpublish <branch>'),
            crayons.red('legit branches [<wildcard pattern>]'),
        )

    help = help.replace('Commands:', additional_help)

    return help


def black(s, **kwargs):
    if legit_settings.allow_black_foreground:
        return crayons.black(s, **kwargs)
    else:
        return s


def git_version():
    """Return the git version tuple (major, minor, patch)"""
    from git import Git

    g = Git()
    return g.version_info


def program_path():
    result = os.path.abspath(sys.argv[0])
    result = result.replace(os.sep, '/').replace(' ', '\\ ')
    return result



# -*- coding: utf-8 -*-

"""
legit.scm
~~~~~~~~~

This module provides the main interface to Git.
"""

import fnmatch
import os
import sys
from collections import namedtuple
from operator import attrgetter

import click
from clint.textui import colored, columns
import crayons
from git import Repo
from git.exc import GitCommandError, InvalidGitRepositoryError

from .settings import legit_settings
from .utils import black, status_log

LEGIT_TEMPLATE = 'Legit: stashing before {0}.'


Branch = namedtuple('Branch', ['name', 'is_published'])


class SCMRepo(object):
    git = None
    repo = None
    remote = None
    verbose = False
    fake = False
    stash_index = None

    def __init__(self):
        self.git = os.environ.get('GIT_PYTHON_GIT_EXECUTABLE', 'git')

        try:
            self.repo = Repo(search_parent_directories=True)
            self.remote = self.get_remote()
        except InvalidGitRepositoryError:
            self.repo = None

    def git_exec(self, command, **kwargs):
        """Execute git commands"""
        from .cli import verbose_echo

        command.insert(0, self.git)
        if kwargs.pop('no_verbose', False):  # used when git output isn't helpful to user
            verbose = False
        else:
            verbose = self.verbose
        verbose_echo(' '.join(command), verbose, self.fake)

        if not self.fake:
            result = self.repo.git.execute(command, **kwargs)
        else:
            if 'with_extended_output' in kwargs:
                result = (0, '', '')
            else:
                result = ''
        return result

    def repo_check(self, require_remote=False):
        if self.repo is None:
            click.echo('Not a git repository.')
            sys.exit(128)

        # TODO: no remote fail
        if not self.repo.remotes and require_remote:
            click.echo('No git remotes configured. Please add one.')
            sys.exit(128)

        # TODO: You're in a merge state.

    def stash_log(self, sync=False):
        if self.repo.is_dirty():
            status_log(self.stash_it, 'Saving local changes.', sync=sync)

    def unstash_log(self, sync=False):
        self.stash_index = self.unstash_index(sync=sync)
        if self.stash_index:
            status_log(self.unstash_it, 'Restoring local changes.', sync=sync)

    def unstash_index(self, sync=False, branch=None):
        """Returns an unstash index if one is available."""

        stash_list = self.git_exec(['stash', 'list'], no_verbose=True)

        if branch is None:
            branch = self.get_current_branch_name()

        for stash in stash_list.splitlines():

            verb = 'syncing' if sync else 'switching'

            if (
                (('Legit' in stash) and
                 ('On {0}:'.format(branch) in stash) and
                 (verb in stash)
                 ) or
                (('GitHub' in stash) and
                 ('On {0}:'.format(branch) in stash) and
                 (verb in stash)
                 )
            ):
                return stash[7]

    def stash_it(self, sync=False):
        msg = 'syncing branch' if sync else 'switching branches'

        return self.git_exec(
            ['stash', 'save', '--include-untracked', LEGIT_TEMPLATE.format(msg)])

    def unstash_it(self, sync=False):
        """
        Unstashes changes from current branch for branch sync.
        Requires prior code setting self.stash_index.
        """
        if self.stash_index is not None:
            return self.git_exec(
                ['stash', 'pop', 'stash@{{{0}}}'.format(self.stash_index)])

    def smart_pull(self):
        """
        'git log --merges origin/master..master'
        """
        branch = self.get_current_branch_name()

        self.git_exec(['fetch', self.remote.name])

        return self.smart_merge('{0}/{1}'.format(self.remote.name, branch),
                                self.smart_merge_enabled())

    def smart_merge_enabled(self):
        reader = self.repo.config_reader()
        if reader.has_option('legit', 'smartMerge'):
            return reader.getboolean('legit', 'smartMerge')
        else:
            return True

    def smart_merge(self, branch, allow_rebase=True):

        from_branch = self.get_current_branch_name()

        merges = self.git_exec(
            ['log', '--merges', '{0}..{1}'.format(branch, from_branch)])

        if allow_rebase:
            verb = 'merge' if merges.count('commit') else 'rebase'
        else:
            if self.pull_rebase():
                verb = 'rebase'
            else:
                verb = 'merge'

        if verb != 'rebase' and self.pull_ff_only():
            return self.git_exec([verb, '--ff-only', branch])
        else:
            try:
                return self.git_exec([verb, branch])
            except GitCommandError as why:
                log = self.git_exec([verb, '--abort'])
                abort('Merge failed. Reverting.',
                      log='{0}\n{1}'.format(why, log), type='merge')

    def pull_rebase(self):
        reader = self.repo.config_reader()
        if reader.has_option('pull', 'rebase'):
            return reader.getboolean('pull', 'rebase')
        else:
            return False

    def pull_ff_only(self):
        reader = self.repo.config_reader()
        if reader.has_option('pull', 'ff'):
            if reader.get('pull', 'ff') == 'only':
                return True
            else:
                return False
        else:
            return False

    def push(self, branch=None):

        if branch is None:
            return self.git_exec(['push'])
        else:
            return self.git_exec(['push', self.remote.name, branch])

    def checkout_branch(self, branch):
        """Checks out given branch."""

        _, stdout, stderr = self.git_exec(
            ['checkout', branch],
            with_extended_output=True)
        return '\n'.join([stderr, stdout])

    def unpublish_branch(self, branch):
        """Unpublishes given branch."""

        try:
            return self.git_exec(
                ['push', self.remote.name, ':{0}'.format(branch)])
        except GitCommandError:
            _, _, log = self.git_exec(
                ['fetch', self.remote.name, '--prune'],
                with_extended_output=True)
            abort('Unpublish failed. Fetching.', log=log, type='unpublish')

    def publish_branch(self, branch):
        """Publishes given branch."""

        return self.git_exec(
            ['push', '-u', self.remote.name, branch])

    def undo(self, hard=False):
        """Makes last commit not exist"""

        if not self.fake:
            return self.repo.git.reset('HEAD^', working_tree=hard)
        else:
            click.echo(crayons.red('Faked! >>> git reset {}{}'
                                   .format('--hard ' if hard else '', 'HEAD^')))
            return 0

    def get_remote(self):

        self.repo_check()

        reader = self.repo.config_reader()

        # If there is no remote option in legit section, return default
        if reader.has_option('legit', 'remote'):
            remote_name = reader.get('legit', 'remote')
            if remote_name not in [r.name for r in self.repo.remotes]:
                if fallback_enabled(reader):
                    return self.get_default_remote()
                else:
                    click.echo('Remote "{0}" does not exist!'.format(remote_name))
                    will_aborted = click.confirm(
                        '\nPress `Y` to abort now,\n' +
                        '`n` to use default remote and turn fallback on for this repo:')
                    if will_aborted:
                        click.echo('\nAborted. Please update your git configuration.')
                        sys.exit(64)  # EX_USAGE
                    else:
                        writer = self.repo.config_writer()
                        writer.set_value('legit', 'remoteFallback', 'true')
                        click.echo('\n`legit.RemoteFallback` changed to true for current repo.')
                        return self.get_default_remote()
            else:
                return self.repo.remote(remote_name)
        else:
            return self.get_default_remote()

    def get_default_remote(self):
        if len(self.repo.remotes) == 0:
            return None
        else:
            return self.repo.remotes[0]

    def get_current_branch_name(self):
        """Returns current branch name"""

        return self.repo.head.ref.name

    def fuzzy_match_branch(self, branch):
        if not branch:
            return False

        all_branches = self.get_branch_names()
        if branch in all_branches:
            return branch

        def branch_fuzzy_match(b):
            return b.startswith(branch)

        possible_branches = list(filter(branch_fuzzy_match, all_branches))

        if len(possible_branches) == 1:
            return possible_branches[0]

        return branch

    def get_branches(self, local=True, remote_branches=True, wildcard_pattern='*'):
        """Returns a list of local and remote branches matching wildcard_pattern."""

        if not self.repo.remotes:
            remote_branches = False

        branches = []

        if remote_branches:

            # Remote refs.
            try:
                for b in self.remote.refs:
                    name = '/'.join(b.name.split('/')[1:])

                    if name not in legit_settings.forbidden_branches:
                        branches.append(Branch(name, is_published=True))
            except (IndexError, AssertionError):
                pass

        if local:

            # Local refs.
            for b in [h.name for h in self.repo.heads]:

                if (not remote_branches) or (b not in [br.name for br in branches]):
                    if b not in legit_settings.forbidden_branches:
                        branches.append(Branch(b, is_published=False))

        matching_branch_names = fnmatch.filter(
            [branch.name for branch in branches], wildcard_pattern
        )
        branches = [branch for branch in branches if branch.name in matching_branch_names]

        return sorted(branches, key=attrgetter('name'))

    def get_branch_names(self, local=True, remote_branches=True):

        branches = self.get_branches(local=local, remote_branches=remote_branches)

        return [b.name for b in branches]

    def display_available_branches(self, wildcard_pattern='*'):
        """Displays available branches."""

        if not self.repo.remotes:
            remote_branches = False
        else:
            remote_branches = True

        branches = self.get_branches(
            local=True, remote_branches=remote_branches, wildcard_pattern=wildcard_pattern
        )
        if not branches:
            click.echo(crayons.red('No branches available'))
            return

        branch_col = len(max([b.name for b in branches], key=len)) + 1

        for branch in branches:

            try:
                branch_is_selected = (branch.name == self.get_current_branch_name())
            except TypeError:
                branch_is_selected = False

            marker = '*' if branch_is_selected else ' '
            color = colored.green if branch_is_selected else colored.yellow
            pub = '(published)' if branch.is_published else '(unpublished)'

            click.echo(columns(
                [colored.red(marker), 2],
                [color(branch.name, bold=True), branch_col],
                [black(pub), 14]
            ))


# Instead of getboolean('legit', 'remoteFallback', fallback=False)
# since getboolean in Python 2 does not have fallback argument.
def fallback_enabled(reader):
    if reader.has_option('legit', 'remoteFallback'):
        return reader.getboolean('legit', 'remoteFallback')
    else:
        return False


class Aborted(object):

    def __init__(self):
        self.message = None
        self.log = None


def abort(message, log=None, type=None):

    a = Aborted()
    a.message = message
    a.log = log

    legit_settings.abort_handler(a, type=type)



# -*- coding: utf-8 -*-

"""
legit.bootstrap
~~~~~~~~~~~~~~~

This module boostraps the Legit runtime.
"""


from clint import resources
from clint.textui import colored
import crayons
from six.moves import configparser

from .settings import legit_settings

resources.init('kennethreitz', 'legit')

try:
    config_file = resources.user.open('config.ini', 'r')
except IOError:
    resources.user.write('config.ini', '')
    config_file = resources.user.open('config.ini', 'r')


# Load existing configuration.
config = configparser.ConfigParser()
try:
    # `read_file()` added in Python 3.2
    config.read_file(config_file)
except AttributeError:
    config.readfp(config_file)


# Populate if needed.
if not config.has_section('legit'):
    config.add_section('legit')


modified = False

# Set defaults if they are missing.
# Add everything to settings object.
for (k, v, _) in legit_settings.config_defaults:
    if not config.has_option('legit', k):
        modified = True
        config.set('legit', k, v)
        setattr(legit_settings, k, v)
    else:
        val = config.get('legit', k)

        # Map boolean strings.
        if val.lower() in ('true', '1', 'yep', 'sure'):
            val = True
        elif val.lower() in ('false', '0', 'nope', 'nadda', 'nah'):
            val = False

        setattr(legit_settings, k, val)

if modified:
    config_file = resources.user.open('config.ini', 'w')
    config.write(config_file)


if legit_settings.disable_colors:
    crayons.disable()
    colored.DISABLE_COLOR = True
